library(plyr)
library(readxl)
library(xlsx)
library(DiffCorr)
library(corrplot)

########################################
########################################
get_lower_tri<-function(cormat)
{
    cormat[upper.tri(cormat)] <- NA
    return(cormat)
}

# Get upper triangle of the correlation matrix
get_upper_tri <- function(cormat)
{
    cormat[lower.tri(cormat)]<- NA
    return(cormat)
}

createLink <- function(val,label) 
{
  sprintf('<a href="%s" target="_blank" class="btn btn-primary">%s</a>',val,label)
}

DisplayWorksheets = function(wkshtList)
{
	df=data.frame()
	for (i in names(wkshtList))
	{
		df=rbind(df,data.frame(wksht=i,value=wkshtList[[i]]))
	}
	return(df)
}

##########################################################
CreateResultDir = function(appName,studyID)
{
	resultDir=sprintf("/var/www/html/shiny/%s_%s",appName,studyID)
	system(sprintf('rm -r %s',resultDir))
	dir.create(resultDir,showWarnings=FALSE)
	
	return(resultDir)
}

##########################################################
GetWrkshtClass = function(inputList)
{
	worksheetClass=list()
	nonNumDataWorksheets=c('ClinicalCategorical','MatchInfo','Group','Confound','corConfound',names(inputList)[grep('_nodeColor$',names(inputList))])
	numDataWorksheets=setdiff(names(inputList),nonNumDataWorksheets)
	worksheetClass[['nonNumDataWorksheets']]=nonNumDataWorksheets
	worksheetClass[['numDataWorksheets']]=numDataWorksheets
	worksheetClass[['analysisWorksheets']]=setdiff(numDataWorksheets,'ClinicalNumeric')
	worksheetClass[['numCatDataWorksheets']]=c(numDataWorksheets,'ClinicalCategorical')
	
	return(worksheetClass)
}

########################################################
# DisplayDf
########################################################
DisplayDf = function(inputList,appName,projectID)
{
	inputDataInfo=data.frame()
	for (i in names(inputList))
	{
		row=nrow(inputList[[i]])
		ncol=ncol(inputList[[i]])
		colNames=''
		if (ncol<10)
		{
			colNames=paste(colnames(inputList[[i]]),collapse=", ")
		}
		inputDataInfo=rbind.fill(inputDataInfo,data.frame(worksheet=i,row=row,col=ncol,colNames=colNames))
	}
	
	inputDataInfo=rbind.fill(inputDataInfo,data.frame(worksheet='resultLink',row=NA,col=NA,colNames=createLink(sprintf("http://13.229.20.244/shiny/%s_%s/",appName,projectID),'Project Result Page')))
	
	return(inputDataInfo)
}

########################################################
# ReadXlsxData
########################################################
ReadXlsxData = function(filename)
{
	worksheets=excel_sheets(filename)
	for (i in worksheets)
	{
		str=paste(i,"=as.data.frame(read_excel('",filename,"',sheet='",i,"'))",sep="")
		eval(parse(text = str))
	}

	########################################################
	# Prepare data
	########################################################
	dataWorksheets=setdiff(worksheets,c('Group','MatchInfo',worksheets[grep('_nodeColor$',worksheets)]))
	
	### add matchinfo to group if matchinfo available
	if (length(intersect(worksheets,'MatchInfo'))==1)
	{
		Group=merge(Group,MatchInfo,by=c('SampleID'),all.x=TRUE)
	}
	
	### Merge Group info into data 
	for (i in dataWorksheets)
	{
		str=paste(i,"=merge(",i,",Group,by=c('SampleID'),all.x=TRUE)",sep="")
		eval(parse(text = str))
		if (length(which(dataWorksheets=='MatchInfo'))>0)
		{
			str=paste(i,"=merge(",i,",MatchInfo,by=c('SampleID'),all.x=TRUE)",sep="")
			eval(parse(text = str))
		}
	}

	####################################################################
	##### Check sample mismatch between trainGroup and trainLipid
	# str=paste("inLipidNotInGroup=setdiff(","Lipid$SampleID,","Group$SampleID)",sep="")
	# eval(parse(text=str))
	# if (length(inLipidNotInGroup)>0)
	# {
		# print(sprintf("SampleIDs in lipid not in group (%d): %s",length(inLipidNotInGroup),paste(inLipidNotInGroup,collapse=",")))
	# }

	# str=paste("inGroupNotInLipid=setdiff(","Group$SampleID,","Lipid$SampleID)",sep="")
	# eval(parse(text=str))
	# if (length(inGroupNotInLipid)>0)
	# {
		# print(sprintf("SampleIDs in group not in lipid (%d): %s",length(inGroupNotInLipid),paste(inGroupNotInLipid,collapse=",")))
	# }
	
	
	
	##############################################################
	# Store all dataframes into list
	dataList=list()
	for (i in worksheets)
	{
		str=paste("dataList[['",i,"']]=",i,sep="")
		eval(parse(text=str))
	}
	
	
	return(dataList)
}


################################################################################
################################################################################
compcorr = function (n1, r1, n2, r2) 
{
    num1a <- which(r1 >= 0.99)
    num2a <- which(r2 >= 0.99)
    r1[num1a] <- 0.99
    r2[num2a] <- 0.99
    num1b <- which(r1 <= -0.99)
    num2b <- which(r2 <= -0.99)
    r1[num1b] <- -0.99
    r2[num2b] <- -0.99
    z1 <- atanh(r1)
    z2 <- atanh(r2)
    dz <- (z1 - z2)/sqrt(1/(n1 - 3) + (1/(n2 - 3)))
    pv <- 2 * (1 - pnorm(abs(dz)))
    return(list(diff = dz, pval = pv))
}

################################################################################
# r to z Fisher's Exact Test
# cc1, cc2 = correlation matrix of group 1 and group 2 respectively
# n1, n2 = sample size of cc1 and cc2 respectively
################################################################################
ComputeDiffCorr = function(cc1,cc2,n1,n2,corrCutoff,pvalCutoff)
{
	#cc1 <- cor(data1, method = method)
	#cc2 <- cor(data2, method = method)
	ccc1 <- as.vector(cc1[lower.tri(cc1)])
	ccc2 <- as.vector(cc2[lower.tri(cc2)])
	n <- nrow(cc1)
	N <- n * (n - 1)/2
	p1 <- rep(1, N)
	p2 <- rep(1, N)
	pdiff <- rep(1, N)
	diff <- rep(1, N)
	mol.names <- rownames(cc1)
	p1 <- cor2.test(n1, ccc1)
	p2 <- cor2.test(n2, ccc2)
	pdiff <- compcorr(n1, ccc1, n2, ccc2)$pval
	diff <- ccc1 - ccc2
	myindex <- which((lower.tri(cc1)) == TRUE, arr.ind = TRUE)
    mol.names1 <- mol.names[myindex[, 2]]
    mol.names2 <- mol.names[myindex[, 1]]
	pdiff.adj=p.adjust(pdiff,method='BH')
    fin.ind <- pdiff.adj < pvalCutoff & ((abs(ccc1)>corrCutoff & abs(ccc2)<corrCutoff) | (abs(ccc1)<corrCutoff & abs(ccc2)>corrCutoff))
    res <- data.frame(cbind(mol.names1[fin.ind], mol.names2[fin.ind], ccc1[fin.ind], p1[fin.ind], ccc2[fin.ind], p2[fin.ind], pdiff[fin.ind], diff[fin.ind],pdiff.adj[fin.ind]))
    head <- c("Var1", "Var2", "r1", "p1", "r2", "p2", 
        "pdiff", "diff","pdiff.adj")
    names(res) <- head
	
	return(res)
}

################################################################################
################################################################################
TestNormality = function(inputX)
{
	iqrsd=as.numeric()
	shapiro=as.numeric()
	for (i in 1:ncol(inputX))
	{
		x=inputX[,i]
		#F_X^{-1}(0.75) - F_X^{-1}(0.25) = 1.34898\sigma,
		sdx=sd(x)
		IQRx=IQR(x)
		
		iqrsd=c(iqrsd,IQRx/sdx)
		
		## Shapiro Test: The null-hypothesis of this test is that the population is normally distributed
		## p < alpha => distribution not normally distributed
		## A Q–Q plot is required for verification in addition to the test.
		shapiro=c(shapiro,shapiro.test(x)$p.value)
	}
	
	
	testResult=data.frame(variable=names(inputX),IQR_sd_Ratio=iqrsd,shapiro_pval=shapiro)
	
	return(testResult)
}

###################################################################################
#### Plot igraph
# PlotIGraph(tmpMatPearson,varList,inputList,uniqueGrp[i],networkDir,corrCutoff,aws)
###################################################################################
PlotIGraph = function(corrMatrix,varList,inputList,uniqueGrpI,networkDir,corrCutoff,aws,networkX,networkY)
{
	corrMatrix<-as.matrix(corrMatrix)
	corrMatrix[corrMatrix< corrCutoff]=0
	diag(corrMatrix)=0
	colnames(corrMatrix)=varList
	rownames(corrMatrix)=varList
	
	graphLipidt0 <- graph.adjacency(corrMatrix, weighted=TRUE, mode="lower")
	if (exists(sprintf("%s_nodeColor",aws),where=inputList))
	{
		colorDf=inputList[[sprintf("%s_nodeColor",aws)]]
		varListDf=data.frame(variable=varList)
		varListDf=merge(varListDf,colorDf,by=c('variable'),all.x=TRUE)
		varListDf$colorGrp[which(is.na(varListDf$colorGrp))]='notAssigned'
		pal <- rainbow(n=length(unique(varListDf$colorGrp)))
		colorMap=setNames(pal,unique(varListDf$colorGrp))
		varListDf$color=colorMap[varListDf$colorGrp]
		varColorMap=setNames(varListDf$color,varListDf$variable)
		
		V(graphLipidt0)$color=varColorMap[varList]
		
		png(sprintf("%s/network_Grp%s_NoLabel.png",networkDir,uniqueGrpI))
		plot(graphLipidt0, vertex.size=6,vertex.label=NA, vertex.color=V(graphLipidt0)$color)
		legend(as.numeric(networkX),as.numeric(networkY), legend=names(colorMap)  , col = pal , bty = "n", pch=20 , pt.cex = 1, cex = 0.6, text.col=pal , horiz = FALSE, inset = c(0.1, 0.1))
		dev.off()
		
		png(sprintf("%s/network_Grp%s_Label.png",networkDir,uniqueGrpI))
		plot(graphLipidt0, vertex.size=6,vertex.label=V(graphLipidt0)$name, vertex.label.cex=0.6, vertex.color=V(graphLipidt0)$color)
		legend(as.numeric(networkX),as.numeric(networkY), legend=names(colorMap)  , col = pal , bty = "n", pch=20 , pt.cex = 1, cex = 0.6, text.col=pal , horiz = FALSE, inset = c(0.1, 0.1))
		dev.off()

	} else
	{
		png(sprintf("%s/network_Grp%s_NoLabel.png",networkDir,uniqueGrpI))
		plot(graphLipidt0, vertex.size=6,vertex.label=NA)
		dev.off()
		
		png(sprintf("%s/network_Grp%s_Label.png",networkDir,uniqueGrpI))
		plot(graphLipidt0, vertex.size=6,vertex.label=V(graphLipidt0)$name, vertex.label.cex=0.6)
		dev.off()
	}
}

################################################################################
################################################################################
DiffCorrNetwork = function(inputList,corrCutoff,diffCorDir,networkX,networkY)
{
	overallDysDf=data.frame()
	normTestOverallDf=data.frame()
	
	corrCutoff=as.numeric(corrCutoff)
	corFilename=sprintf("%s/corr_file.xlsx",diffCorDir)
	if (file.exists(corFilename)) file.remove(corFilename)
	
	worksheets=GetWrkshtClass(inputList)
	analysisWorksheets=worksheets[['analysisWorksheets']]
	
	
	for (aws in analysisWorksheets)
	{
		fullDysRegVarList=as.character()
		pcorAwsDir=sprintf("%s/%s",diffCorDir,aws)
		dir.create(pcorAwsDir,showWarnings=FALSE)
		
		networkDir=sprintf("%s/%s/network",diffCorDir,aws)
		dir.create(networkDir,showWarnings=FALSE)
		
		inputData=inputList[[aws]]
		uniqueGrp=unique(inputData$GroupID)
		
		varList=setdiff(names(inputData),c('SampleID','GroupID','MatchID'))
		varListComb=t(combn(seq(1:length(varList)),2))
		
		corCoef=list()
		
		corr_method='Correlation'
		if (exists("corConfound",where=inputList)) ### If confounders to be considered, perform partial correlation
		{
			corr_method='Partial Correlation'
			corConfounders=inputList[['corConfound']]
			confoundList=setdiff(names(corConfounders),c('SampleID','GroupID','MatchID'))
			inputData=merge(inputData,corConfounders[,c('SampleID',confoundList)],by=c('SampleID'),all.x=TRUE)

			#### Compute pcor for all unique groups
			for (i in 1:length(uniqueGrp))
			{
				tmpData=inputData[which(inputData$GroupID==uniqueGrp[i]),]
				
				regressTmp=matrix(nrow=nrow(tmpData),ncol=length(varList))
				for (j in 1:length(varList))
				{
					xData=tmpData[,c(varList[j],confoundList)]
					names(xData)[1]='X'
					regress_x = lm(X~.,data=xData)
					residuals_x = residuals(regress_x)
					regressTmp[,j]=residuals_x
				}
				regressTmp = data.frame(regressTmp)
				names(regressTmp)=varList
				normTestResult=TestNormality(regressTmp[,varList])
				normTestResult$GroupID=uniqueGrp[i]
				normTestResult$worksheet=aws
				tmpMatPearson=cor(regressTmp,method='pearson')
				tmpMatSpearman=cor(regressTmp,method='spearman')
				corCoef[[sprintf("%s_pearson",i)]]=tmpMatPearson
				corCoef[[sprintf("%s_spearman",i)]]=tmpMatSpearman
				
				normTestOverallDf=rbind(normTestOverallDf,normTestResult)
				
				#### Plot igraph
				PlotIGraph(tmpMatPearson,varList,inputList,uniqueGrp[i],networkDir,corrCutoff,aws,networkX,networkY)
				
			}
		} else ### if no confounders, perform correlation
		{
			for (i in 1:length(uniqueGrp))
			{
				tmpData=inputData[which(inputData$GroupID==uniqueGrp[i]),]
				
				tmpMatPearson=cor(tmpData[,varList],method='pearson')
				tmpMatSpearman=cor(tmpData[,varList],method='spearman')
				corCoef[[sprintf("%s_pearson",i)]]=tmpMatPearson
				corCoef[[sprintf("%s_spearman",i)]]=tmpMatSpearman
				
				normTestResult=TestNormality(tmpData[,varList])
				normTestResult$GroupID=uniqueGrp[i]
				normTestResult$worksheet=aws
				normTestOverallDf=rbind(normTestOverallDf,normTestResult)
				
				PlotIGraph(tmpMatPearson,varList,inputList,uniqueGrp[i],networkDir,corrCutoff,aws,networkX,networkY)
			}
		}
		
		
		# For each combination of unique groups, plot heatmap
		ncomb=t(combn(seq(1,length(uniqueGrp)),2))
		
		splotDir=sprintf("%s/corrScatterPlot",pcorAwsDir)
		dir.create(splotDir, showWarnings=FALSE)
		
		for (i in 1:nrow(ncomb))
		{
			for (method in c("pearson","spearman"))
			{
				corr1=corCoef[[sprintf("%s_%s",ncomb[i,1],method)]]
				corr2=corCoef[[sprintf("%s_%s",ncomb[i,2],method)]]
				
				n1=length(which(inputData$GroupID==uniqueGrp[ncomb[i,1]]))
				n2=length(which(inputData$GroupID==uniqueGrp[ncomb[i,2]]))
				diffCorrResult=ComputeDiffCorr(corr1,corr2,n1,n2,corrCutoff,0.05)
				
				if (nrow(diffCorrResult)>0)
				{
					diffCorrResult$worksheet=aws
					diffCorrResult$Group1=uniqueGrp[ncomb[i,1]]
					diffCorrResult$Group2=uniqueGrp[ncomb[i,2]]
					diffCorrResult$method=method
					overallDysDf=rbind(overallDysDf,diffCorrResult)
					
					### Plot pcor plots for pair
					dysRegVarList=unique(c(as.character(diffCorrResult$Var1),as.character(diffCorrResult$Var2)))
					fullDysRegVarList=unique(c(fullDysRegVarList,dysRegVarList))
					
					png(sprintf("%s/pcorPlot_grp%s_grp%s_%s.png",pcorAwsDir,uniqueGrp[ncomb[i,1]],uniqueGrp[ncomb[i,2]],method))
					par(mfrow=c(1,2))
					corrplot(corr1[dysRegVarList,dysRegVarList], type="lower", tl.cex=0.5,title=sprintf("%s (%s) for Grp %d",corr_method,method,uniqueGrp[ncomb[i,1]]))
					corrplot(corr2[dysRegVarList,dysRegVarList], type="lower", tl.cex=0.5,title=sprintf("%s (%s) for Grp %d",corr_method,method,uniqueGrp[ncomb[i,2]]))
					dev.off()
					
					###### Plot scatter plot
					for (k in 1:nrow(diffCorrResult))
					{
						var1=as.character(diffCorrResult$Var1[k])
						var2=as.character(diffCorrResult$Var2[k])
						r1=as.numeric(as.character(diffCorrResult$r1[k]))
						r2=as.numeric(as.character(diffCorrResult$r2[k]))
						png(sprintf('%s/%s_vs_%s_%s.png',splotDir,gsub("[\\./:()]","-",var1),gsub("[\\./:()]","-",var2),method))
						par(mfrow=c(1,2))
						x=inputData[which(inputData$GroupID==uniqueGrp[ncomb[i,1]]),var1]
						y=inputData[which(inputData$GroupID==uniqueGrp[ncomb[i,1]]),var2]
						plot(x,y,xlab=sprintf("%s",var1),ylab=sprintf("%s",var2),main=sprintf('Grp %s r=%0.2f',uniqueGrp[ncomb[i,1]],r1))
						abline(lm(y ~ x),col='red')
						x=inputData[which(inputData$GroupID==uniqueGrp[ncomb[i,2]]),var1]
						y=inputData[which(inputData$GroupID==uniqueGrp[ncomb[i,2]]),var2]
						plot(x,y,xlab=sprintf("%s",var1),ylab=sprintf("%s",var2),main=sprintf('Grp %s r=%0.2f',uniqueGrp[ncomb[i,2]],r2))
						abline(lm(y ~ x),col='red')
						dev.off()
					}
				}
				
				
			}
			
		}
		
		# Plot heatmap for all groups (if more than 2 groups else already plotted in above loop)
		if (length(uniqueGrp)>2)
		{
			
			for (method in c('pearson','spearman'))
			{
				png(sprintf("%s/pcorPlot_%s_All_%s.png",pcorAwsDir,studyID,method))
				par(mfrow=c(1,length(uniqueGrp)))
				for (i in 1:length(uniqueGrp))
				{
					corrTmp=corCoef[[sprintf("%s_%s",ncomb[i,1],method)]]
					corrplot(corrTmp[fullDysRegVarList,fullDysRegVarList], type="lower", tl.cex=0.5)
				}
				dev.off()
			}
		}
	}
	
	write.xlsx(overallDysDf,corFilename,sheetName='overallDysDf',append=TRUE)
	write.xlsx(normTestOverallDf,corFilename,sheetName='normalityTest',append=TRUE)
	
	return(overallDysDf)
}







